

#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2017-2020 - The Aura Engine Team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.



import logging

from datetime import datetime


from src.base.config import AuraConfig
from src.base.utils import SimpleUtil as SU
from src.core.engine import Engine
from src.scheduling.utils import M3UPlaylistProcessor
from src.scheduling.models import Timeslot, Playlist, PlaylistEntry, PlaylistEntryMetaData
from src.scheduling.api import ApiFetcher



class ProgrammeService():
    """
    The current programme of the calendar. The programme is a set of timeslots for the current day.
    """
    config = None
    logger = None
    timeslots = None
    api_fetcher = None
    last_successful_fetch = None
    programme_store = None


    def __init__(self):
        """
        Constructor
        """
        self.config = AuraConfig.config()
        self.logger = logging.getLogger("AuraEngine")    
        self.programme_store = ProgrammeStore()



    def refresh(self):
        """
        Fetch the latest programme from `ProgrammeStore` which stores it to the database.
        After that, the programme is in turn loaded from the database and stored in `self.timeslots`.
        """

        # Fetch programme from API endpoints
        self.logger.debug("Trying to fetch new programme from API endpoints...")

        # Create a fetching thread and wait until it's done
        self.api_fetcher = ApiFetcher(self.config)
        self.api_fetcher.start()
        response = self.api_fetcher.get_fetched_data()

        if response is None:
            msg = SU.red("Trying to load programme from Engine Database, because ApiFetcher returned an empty response.")
            self.logger.warning(msg)
        elif type(response) is list:
            if len(response) > 0:
                self.last_successful_fetch = datetime.now()
                self.timeslots = self.programme_store.store_timeslots(response)
                self.logger.info(SU.green(f"Finished fetching current programme from API ({len(response)} timeslots)"))
            else:
                self.logger.critical("Programme fetched from Steering/Tank has no entries!")
        elif response.startswith("fetching_aborted"):
            msg = SU.red("Load programme from DB, because fetching was aborted by ApiFetcher! Reason: " + response[17:])
            self.logger.warning(msg)
        else:
            msg = SU.red("Load programme from DB, because of an unknown response from ApiFetcher: " + response)
            self.logger.warning(msg)

        # Load latest programme from the database        
        if not self.timeslots:
            self.timeslots = self.programme_store.load_timeslots()
            self.logger.info(SU.green("Finished loading current programme from database (%s timeslots)" % str(len(self.timeslots))))

        for timeslot in self.timeslots:
            self.logger.debug("\tTimeslot %s with Playlist %s" % (str(timeslot), str(timeslot.playlist)))



    def get_current_entry(self):
        """
        Retrieves the current `PlaylistEntry` which should be played as per programme. 

        Returns:
            (PlaylistEntry): The track which is (or should) currently being played
        """
        now_unix = Engine.engine_time()

        # Load programme if necessary
        if not self.timeslots:
            self.refresh()

        # Check for current timeslot
        current_timeslot = self.get_current_timeslot()
        if not current_timeslot:
            self.logger.warning(SU.red("There's no active timeslot"))
            return None

        # Check for scheduled playlist
        current_playlist = self.get_current_playlist(current_timeslot)
        if not current_playlist:
            msg = "There's no (default) playlist assigned to the current timeslot. Most likely a fallback will make things okay again."
            self.logger.warning(SU.red(msg))
            return None

        # Iterate over playlist entries and store the current one
        current_entry = None
        for entry in current_playlist.entries:
            if entry.start_unix <= now_unix and now_unix <= entry.end_unix:
                current_entry = entry
                break
      
        if not current_entry:
            # Nothing playing ... fallback will kick-in
            msg = "There's no entry scheduled for playlist '%s' at %s" % (str(current_playlist), SU.fmt_time(now_unix))
            self.logger.warning(SU.red(msg))
            return None

        return current_entry
 


    def get_current_timeslot(self):
        """
        Retrieves the timeslot currently to be played. 
        
        Returns:
            (Timeslot): The current timeslot
        """
        current_timeslot = None
        now_unix = Engine.engine_time()

        # Iterate over all timeslots and find the one to be played right now
        if self.timeslots:
            for timeslot in self.timeslots:
                if timeslot.start_unix <= now_unix and now_unix < timeslot.end_unix:
                    current_timeslot = timeslot
                    break
        
        return current_timeslot



    def get_current_playlist(self, timeslot):
        """
        Retrieves the playlist to be scheduled. If no specific playlist is assigned,
        the default schedule or show playlist is returned. This method does not
        respect any defined fallback playlists.

        Returns:
            (FallbackType, Playlist): The currently assigned playlist
        """        
        playlist = timeslot.playlist
        if not playlist:
            playlist = timeslot.default_schedule_playlist
            if not playlist:
                playlist = timeslot.default_show_playlist

        return playlist



    def get_next_timeslots(self, max_count=0):
        """
        Retrieves the timeslots to be played after the current one.

        Args:
            max_count (Integer): Maximum of timeslots to return, if `0` all exitsing ones are returned

        Returns:
            ([Timeslot]): The next timeslots
        """
        now_unix = Engine.engine_time()
        next_timeslots = []

        if not self.timeslots:
            return []

        for timeslot in self.timeslots:
            if timeslot.start_unix > now_unix:
                if (len(next_timeslots) < max_count) or max_count == 0:
                    next_timeslots.append(timeslot)
                else:
                    break
        
        return next_timeslots



    def is_timeslot_in_window(self, timeslot):
        """
        Checks if the timeslot is within the scheduling window.
        """
        now_unix = Engine.engine_time()
        window_start = self.config.get("scheduling_window_start")
        window_end = self.config.get("scheduling_window_end")

        if timeslot.start_unix - window_start < now_unix and \
            timeslot.start_unix - window_end > now_unix:

            return True
        return False



    def terminate(self):
        """
        Called when thread is stopped or a signal to terminate is received.
        """
        self.logger.info("Shutting down programme service ...")
        if self.api_fetcher:
            self.api_fetcher.terminate()




class ProgrammeStore():
    """ 
    The `ProgrammeStore` service retrieves all current schedules and related
    playlists including audio files from the configured API endpoints and stores 
    it in the local database.

    To perform the API queries it utilizes the ApiFetcher class.
    """
    config = None
    logger = None
    m3u_processor = None

    def __init__(self):
        """
        Initializes the class.
        """
        self.config = AuraConfig.config()
        self.logger = logging.getLogger("AuraEngine")
        self.m3u_processor = M3UPlaylistProcessor()


    def load_timeslots(self):
        """
        Loads the programme from the database.
        """
        try:
            timeslots = Timeslot.get_timeslots(datetime.now())
        except Exception as e:
            self.logger.critical(SU.red("Could not load programme from database. We are in big trouble my friend!"), e)
        return timeslots



    def store_timeslots(self, fetched_timeslots):
        """
        Stores the fetched timeslots to the database.
        """
        timeslots = []

        # Check if existing timeslots have been deleted
        self.update_deleted_timeslots(fetched_timeslots)

        # Process fetched timeslots    
        for timeslot in fetched_timeslots:

            # Check timeslot for validity
            if "start" not in timeslot:
                self.logger.warning("No 'start' of timeslot given. Skipping the timeslot: %s " % str(timeslot))
                continue
            if "end" not in timeslot:
                self.logger.warning("No 'end' of timeslot given. Skipping the timeslot: %s " % str(timeslot))
                continue

            # Store the timeslot
            timeslot_db = self.store_timeslot(timeslot)
            timeslots.append(timeslot_db)

            # Store assigned playlists
            self.store_playlist(timeslot_db, timeslot_db.playlist_id, timeslot["playlist"])
            if timeslot_db.default_schedule_playlist_id:
                self.store_playlist(timeslot_db, timeslot_db.default_schedule_playlist_id, timeslot["default_schedule_playlist"])
            if timeslot_db.default_show_playlist_id:
                self.store_playlist(timeslot_db, timeslot_db.default_show_playlist_id, timeslot["default_show_playlist"])
            if timeslot_db.schedule_fallback_id:
                self.store_playlist(timeslot_db, timeslot_db.schedule_fallback_id, timeslot["schedule_fallback"])
            if timeslot_db.show_fallback_id:
                self.store_playlist(timeslot_db, timeslot_db.show_fallback_id, timeslot["show_fallback"])
            if timeslot_db.station_fallback_id:
                self.store_playlist(timeslot_db, timeslot_db.station_fallback_id, timeslot["station_fallback"])                    

        return timeslots



    def update_deleted_timeslots(self, fetched_timeslots):
        """
        Checks if some timeslot has been deleted remotely, so delete it locally too.
        
        Attention: This method has no effect if only a single timeslot got deleted, 
        because this could simply indicate a issue with the API/Steering, since that 
        means no data got retrieved.

        Args:
            fetched_timeslots ([dict]): List of timeslot dictionaries from the API
        """
        now_unix = SU.timestamp()
        scheduling_window_start = self.config.get("scheduling_window_start")

        local_timeslots = Timeslot.get_timeslots(datetime.now())
        for local_timeslot in local_timeslots:

            # Ignore timeslots which have already started
            if local_timeslot.start_unix > now_unix:

                # Filter the local timeslot from the fetched ones
                existing_remotely = list(filter(lambda new_timeslot: \
                    new_timeslot["timeslot_id"] == local_timeslot.timeslot_id, fetched_timeslots))
                
                if not existing_remotely:
                    # Only allow deletion of timeslots which are deleted before the start of the scheduling window
                    if (local_timeslot.start_unix - scheduling_window_start) > now_unix:                            
                        self.logger.info("Timeslot #%s has been deleted remotely, hence also delete it locally too [%s]" % \
                            (local_timeslot.timeslot_id, str(local_timeslot)))
                        local_timeslot.delete(commit=True)
                        self.logger.info("Remotely deleted timeslot #%s from local database" % local_timeslot.timeslot_id)
                    else:
                        msg = "Timeslot #%s has been deleted remotely. Since the scheduling window has already started, it won't be deleted locally." % local_timeslot.timeslot_id
                        self.logger.error(SU.red(msg))



    def store_timeslot(self, timeslot):
        """
        Stores the given timeslot to the database.

        Args:
            timeslot (Timeslot):    The timeslot
        """
        timeslot_db = Timeslot.for_datetime(timeslot["start"])
        havetoadd = False

        if not timeslot_db:
            self.logger.debug("no timeslot with given timeslot id in database => create new")
            timeslot_db = Timeslot()
            havetoadd = True

        timeslot_db.show_id = timeslot["show_id"]
        timeslot_db.timeslot_id = timeslot["timeslot_id"]
        timeslot_db.timeslot_start = timeslot["start"]
        timeslot_db.timeslot_end = timeslot["end"]
        timeslot_db.show_name = timeslot["show_name"]
        timeslot_db.show_hosts = timeslot["show_hosts"]
        timeslot_db.is_repetition = timeslot["is_repetition"]
        timeslot_db.funding_category = timeslot["show_fundingcategory"]
        timeslot_db.languages = timeslot["show_languages"]
        timeslot_db.type = timeslot["show_type"]
        timeslot_db.category = timeslot["show_categories"]
        timeslot_db.topic = timeslot["show_topics"]
        timeslot_db.musicfocus = timeslot["show_musicfocus"]
        timeslot_db.playlist_id = timeslot["playlist_id"]

        # Optional API properties
        if "default_schedule_playlist_id" in timeslot:
            timeslot_db.default_schedule_playlist_id = timeslot["default_schedule_playlist_id"]
        if "default_show_playlist_id" in timeslot:            
            timeslot_db.default_show_playlist_id = timeslot["default_show_playlist_id"]
        if "schedule_fallback_id" in timeslot:
            timeslot_db.schedule_fallback_id = timeslot["schedule_fallback_id"]
        if "show_fallback_id" in timeslot:
            timeslot_db.show_fallback_id = timeslot["show_fallback_id"]
        if "station_fallback_id" in timeslot:
            timeslot_db.station_fallback_id = timeslot["station_fallback_id"]

        timeslot_db.store(add=havetoadd, commit=True)
        return timeslot_db



    def store_playlist(self, timeslot_db, playlist_id, fetched_playlist):
        """
        Stores the Playlist to the database.
        """
        if not playlist_id or not fetched_playlist:
            self.logger.debug(f"Playlist ID#{playlist_id} is not available!")
            return
        
        playlist_db = Playlist.select_playlist_for_timeslot(timeslot_db.timeslot_start, playlist_id)
        havetoadd = False

        if not playlist_db:
            playlist_db = Playlist()
            havetoadd = True

        self.logger.debug("Storing playlist %d for timeslot (%s)" % (playlist_id, str(timeslot_db)))
        playlist_db.playlist_id = playlist_id
        playlist_db.timeslot_start = timeslot_db.timeslot_start
        playlist_db.show_name = timeslot_db.show_name
        if "entries" in fetched_playlist:
            playlist_db.entry_count = len(fetched_playlist["entries"])
        else:
            playlist_db.entry_count = 0

        playlist_db.store(havetoadd, commit=True)
      
        if playlist_db.entry_count > 0:
            self.store_playlist_entries(timeslot_db, playlist_db, fetched_playlist)

        return playlist_db



    def store_playlist_entries(self, timeslot_db, playlist_db, fetched_playlist):
        """
        Stores the playlist entries to the database.
        """
        entry_num = 0
        time_marker = playlist_db.start_unix
        entries = fetched_playlist["entries"]

        # "Hidden Functionality" to feed engine with M3U playlists via Tank's "Stream" playlist entry type
        # See https://gitlab.servus.at/aura/engine/-/issues/53
        # In the future this is to be replaced by generic music pool feature.
        entries = self.m3u_processor.spread(entries)

        self.expand_entry_duration(timeslot_db, entries)        
        self.delete_orphaned_entries(playlist_db, entries)  

        for entry in entries:
            entry_db = PlaylistEntry.select_playlistentry_for_playlist(playlist_db.artificial_id, entry_num)
            havetoadd = False
            if not entry_db:
                entry_db = PlaylistEntry()
                havetoadd = True

            entry_db.entry_start = datetime.fromtimestamp(time_marker)
            entry_db.artificial_playlist_id = playlist_db.artificial_id
            entry_db.entry_num = entry_num
            entry_db.duration = SU.nano_to_seconds(entry["duration"])

            if "uri" in entry:
                entry_db.source = entry["uri"]
            if "filename" in entry:
                entry_db.source = entry["filename"]

            entry_db.store(havetoadd, commit=True)

            if "file" in entry:
                self.store_playlist_entry_metadata(entry_db, entry["file"]["metadata"])

            entry_num = entry_num + 1
            time_marker += entry_db.duration



    def delete_orphaned_entries(self, playlist_db, entries):
        """
        Deletes all playlist entries which are beyond the current playlist's `entry_count`.
        Such entries might be existing due to a remotely changed playlist, which now has
        less entries than before.
        """
        new_last_idx = len(entries)
        existing_last_idx = PlaylistEntry.count_entries(playlist_db.artificial_id)-1

        if existing_last_idx < new_last_idx:
            return 

        for entry_num in range(new_last_idx, existing_last_idx+1, 1):
            PlaylistEntry.delete_entry(playlist_db.artificial_id, entry_num)            
            self.logger.info(SU.yellow("Deleted playlist entry %s:%s" % (playlist_db.artificial_id, entry_num)))
            entry_num += 1



    def expand_entry_duration(self, timeslot_db, entries):
        """
        If some playlist entry doesn't have a duration assigned, its duration is expanded to the
        remaining duration of the playlist (= timeslot duration minus playlist entries with duration).
        If there's more than one entry without duration, such entries are removed from the playlist.
        """
        total_seconds = (timeslot_db.timeslot_end - timeslot_db.timeslot_start).total_seconds()
        total_duration = SU.seconds_to_nano(total_seconds)
        actual_duration = 0        
        missing_duration = []
        idx = 0

        for entry in entries:
            if not "duration" in entry:
                missing_duration.append(idx)
            else:
                actual_duration += entry["duration"]
            idx += 1
                
        if len(missing_duration) == 1:
            entries[missing_duration[0]]["duration"] = total_duration - actual_duration
            self.logger.info(f"Expanded duration of playlist entry #{missing_duration[0]}")

        elif len(missing_duration) > 1:
            # This case should actually never happen, as TANK doesn't allow more than one entry w/o duration anymore
            for i in reversed(missing_duration[1:-1]):
                self.logger.error(SU.red("Deleted Playlist Entry without duration: %s" % \
                    str(entries[i])))
                del entries[i]



    def store_playlist_entry_metadata(self, entry_db, metadata):
        """
        Stores the meta-data for a PlaylistEntry.
        """
        metadata_db = PlaylistEntryMetaData.select_metadata_for_entry(entry_db.artificial_id)
        havetoadd = False
        if not metadata_db:
            metadata_db = PlaylistEntryMetaData()
            havetoadd = True

        metadata_db.artificial_entry_id = entry_db.artificial_id

        if "artist" in metadata:
            metadata_db.artist = metadata["artist"]
        else:
            metadata_db.artist = ""
        
        if "album" in metadata:
            metadata_db.album = metadata["album"]
        else:
            metadata_db.album = ""

        if "title" in metadata:
            metadata_db.title = metadata["title"]
        else:
            metadata_db.title = ""

        metadata_db.store(havetoadd, commit=True)



